/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.cache.lib;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.cache.api.CacheCapacityEvent;
import org.objectweb.dream.cache.api.CacheCapacityEventListener;
import org.objectweb.dream.cache.api.CacheEntry;
import org.objectweb.dream.cache.api.CacheEntryFactory;
import org.objectweb.dream.cache.api.CacheEntryFilter;
import org.objectweb.dream.cache.api.CacheEvent;
import org.objectweb.dream.cache.api.CacheEventListener;
import org.objectweb.dream.cache.api.CacheException;
import org.objectweb.dream.cache.api.CacheManager;
import org.objectweb.dream.cache.api.FixableCacheEntry;
import org.objectweb.dream.cache.api.InvalidCacheEntryException;
import org.objectweb.dream.cache.api.UnbindManager;
import org.objectweb.dream.cache.replacement.api.ReplacementManager;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Cardinality;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;

/**
 * This class is an implementation of a cache. The data structure is a Hahsed
 * map between cache entry identifier and an entry instance. This class is a
 * fractal component.
 * 
 * @author S.Chassande-Barrioz, E.Bruneton
 * @author P.Garcia
 */
@DreamComponent
public class BasicCacheManager implements CacheManager, UnbindManager,
        BasicCacheManagerAttributeController {

    public final static String DEFAULT_AUTO_CLEAN_SIZE = "7%";

    // ------------------------------------------------------------------------
    // --
    // Attribute
    // ------------------------------------------------------------------------
    // --

    /**
     * This field is the max size of the cache.
     */
    @Attribute(argument = "cacheMaxSize")
    protected int cacheMaxSize;

    /**
     * Threshold of the cache in percentage.
     */
    @Attribute(argument = "autoCleanThreshold")
    protected String autoCleanThresHold;

    // ------------------------------------------------------------------------
    // --
    // Client Interfaces
    // ------------------------------------------------------------------------
    // --

    public final static String CACHE_ENTRY_FACTORY_BINDING = "cache-entry-factory";

    public final static String REPLACEMENT_MANAGER_BINDING = "replacement-manager";

    public final static String CACHE_LISTENER_BINDING = "cache-listeners";

    public final static String CACHE_CAPACTITY_LISTENER_BINDING = "cache-capacity-listeners";

    /**
     * This field is the factory of CacheEntry instances.
     */
    @Requires(name = CACHE_ENTRY_FACTORY_BINDING)
    protected CacheEntryFactory cef;

    /**
     * Replacement Manager : FIFO, LRU, MRU
     */
    @Requires(name = REPLACEMENT_MANAGER_BINDING)
    protected ReplacementManager replacementManager;

    /**
     * This field is the list of CacheEventListener registered.
     */
    @Requires(name = CACHE_LISTENER_BINDING, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    protected Map<String, CacheEventListener> cels;

    /**
     * This field is the list of CacheEventListener registered.
     */
    @Requires(name = CACHE_CAPACTITY_LISTENER_BINDING, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
    protected Map<String, CacheCapacityEventListener> ccels;

    // ------------------------------------------------------------------------
    // --
    // Service
    // ------------------------------------------------------------------------
    // --

    /**
     * Reference to this component
     */
    @Service
    Component ref;

    /**
     * Logger of the component
     */
    @DreamMonolog
    protected Logger logger;

    // ------------------------------------------------------------------------
    // --
    // Fields
    // ------------------------------------------------------------------------
    // --

    /**
     * This field is the cache. This is a Map which the key are object
     * identifiers and the values are FixableCacheEntry instances.
     */
    protected Map<Object, WeakCacheEntry> cache;

    /**
     * Queue used to register the weak references stored in the cache.
     */
    protected ReferenceQueue<CacheEntry> queue;

    /**
     * Thread responsible of cache cleaning
     */
    private BackgroundCleaner bgcleaner;

    /**
     * Threshold of the cache : number of elements. This field is related to the
     * {@link #autoCleanThresHold} attribute
     */
    protected int cacheThresHold = 0;

    private String autoCleanSizeStr;

    private int autoCleanSize;

    private boolean started = false;

    /**
     * It builds a BasicCacheManager with an empty CacheEventListener list, an
     * empty cache and the default max size.
     */
    public BasicCacheManager() {
        this.cels = new HashMap<String, CacheEventListener>();
        this.ccels = new HashMap<String, CacheCapacityEventListener>();
        this.cache = new HashMap<Object, WeakCacheEntry>();
        this.queue = new ReferenceQueue<CacheEntry>();
        this.cacheMaxSize = NO_LIMIT;
        this.started = false;
        this.autoCleanSizeStr = null;
        this.cacheThresHold = 0;
    }

    @LifeCycle(on = LifeCycleType.START)
    public void startFc() throws IllegalLifeCycleException {
        try {
            if (!this.started) {
                this.started = true;
                this.autoCleanSize = this.getValue(this.autoCleanSizeStr);
                this.cacheThresHold = this.getValue(this.autoCleanThresHold);
            }

            final CacheCapacityEvent ev = new CapacityEvent(CacheCapacityEvent.NOTIFY_CACHE_RESIZE,
                    NO_LIMIT, this.cacheMaxSize);
            for (final CacheCapacityEventListener cacheCapacityEventListener : this.ccels.values()) {
                cacheCapacityEventListener.cacheResized(ev);
            }

        } catch (final CacheException e) {
            throw new IllegalLifeCycleException("An error occured while starting the "
                    + BasicCacheManager.class.getSimpleName() + " component :\n " + e.getMessage());
        }
    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stopFc() {
        if (this.started) {
            this.started = false;
            final int maxObjects = this.cacheMaxSize;
            try {
                // resizes the cache to 0 in order to clear it.
                this.setCacheMaxSize(0);
            } catch (final CacheException e) {
                this.logger.log(BasicLevel.WARN, "The Cache manager didn't stop correctly : "
                        + e.getMessage());
            }

            // stop the BackgroundCleaner
            if (this.bgcleaner != null) {
                this.bgcleaner.stop();
                this.bgcleaner = null;
            }

            // restore the value for the future startFc()
            this.cacheMaxSize = maxObjects;
        }
    }

    // IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    // -----------------------------------------------------//

    public void unbindFc(String s) throws NoSuchInterfaceException {
        if (CACHE_ENTRY_FACTORY_BINDING.equals(s)) {
            this.cef = null;
        } else if (REPLACEMENT_MANAGER_BINDING.equals(s)) {
            this.replacementManager = null;
            if (this.bgcleaner != null) {
                this.bgcleaner.stop();
                this.bgcleaner = null;
            }
        } else if (s.startsWith(CACHE_LISTENER_BINDING)) {
            this.cels.remove(s);
        } else if (s.startsWith(CACHE_CAPACTITY_LISTENER_BINDING)) {
            this.ccels.remove(s);
        } else {
            throw new NoSuchInterfaceException(s);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * setCacheMaxSize(int)
     */
    public void setCacheMaxSize(int size) throws IllegalArgumentException, CacheException {
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "resize the cache: "
                    + (size == NO_LIMIT ? "UNLIMITED" : "" + size));
        }

        if (size < 0) {
            throw new IllegalArgumentException(
                    "Impossible to set the cache size to a negative value: " + size);
        }

        if (size == NO_LIMIT) {
            this.cacheMaxSize = NO_LIMIT;
            return;
        }

        synchronized (this.cache) {
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG, "resize the cache: " + this.cacheMaxSize + " to "
                        + size + " (Number of elements in the cache :" + this.cache.size() + ")");
            }

            if (this.started) {
                // if this component is stopping you cannot access to the
                // replacement manager
                int delta = size - this.cache.size();
                if (delta < 0) {
                    this.logger.log(BasicLevel.DEBUG, "free the cache forced");
                    delta = -delta;
                    final int free = this.replacementManager.forceFree(delta);
                    if (delta > free) {
                        this.logger.log(BasicLevel.WARN, "It stays " + (delta - free)
                                + " entries which have not been freed");
                    }
                    this.logger.log(BasicLevel.DEBUG, "cache freed");
                }
            }
            this.cacheMaxSize = size;
            if (this.started) {
                final CacheCapacityEvent ev = new CapacityEvent(
                        CacheCapacityEvent.NOTIFY_CACHE_RESIZE, this.cacheMaxSize, size);
                for (final CacheCapacityEventListener cacheCapacityEventListener : this.ccels
                        .values()) {
                    cacheCapacityEventListener.cacheResized(ev);
                }
            } else {
                // the component has been stopped, rehash the data structure
                final Map<Object, WeakCacheEntry> old = this.cache;
                this.cache = new HashMap<Object, WeakCacheEntry>(this.cacheMaxSize, 1);
                if (old.size() > 0) {
                    this.cache.putAll(old);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getCacheMaxSize()
     */
    public int getCacheMaxSize() {
        return this.cacheMaxSize;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getCurrentSize()
     */
    public int getCurrentSize() {
        return this.cache.size();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getCurrentEntryIdentifiers()
     */
    public Collection<Object> getCurrentEntryIdentifiers() {
        synchronized (this.cache) {
            if (this.cache.size() == 0) {
                return Collections.emptySet();
            }
            return Collections.unmodifiableCollection(this.cache.keySet());
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * setMemorySize(int)
     */
    public void setMemorySize(int size) throws IllegalArgumentException {
        if (size < 0) {
            throw new IllegalArgumentException(
                    "Impossible to set the cache size to a negative value: " + size);
        }
        throw new NotImplementedException();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getMemorySize()
     */
    public int getMemorySize() {
        throw new NotImplementedException();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getCurrentMemorySize()
     */
    public int getCurrentMemorySize() {
        throw new NotImplementedException();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * setAutoCleanSize(java.lang.String)
     */
    public void setAutoCleanSize(String size) throws CacheException {
        this.autoCleanSizeStr = size;
        this.autoCleanSize = this.getValue(this.autoCleanSizeStr);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getAutoCleanSize()
     */
    public String getAutoCleanSize() {
        return this.autoCleanSizeStr;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * setAutoCleanThreshold(java.lang.String)
     */
    public void setAutoCleanThreshold(String size) throws CacheException {
        this.autoCleanThresHold = size;
        this.cacheThresHold = this.getValue(this.autoCleanThresHold);
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.lib.BasicCacheManagerAttributeController#
     * getAutoCleanThreshold()
     */
    public String getAutoCleanThreshold() {
        return this.autoCleanThresHold;
    }

    /**
     * Get the percentage value.
     * 
     * @param str
     * @return
     * @throws CacheException
     */
    private int getValue(String str) throws CacheException {
        if (this.cacheMaxSize == NO_LIMIT) {
            return 0;
        }
        if (str == null) {
            str = "75%";
        }
        final int i = str.indexOf('%');
        if (i == -1) {
            return Integer.parseInt(str);
        } else {
            final int res = Integer.parseInt(str.substring(0, i));
            if (res < 0 || res > 100)
                throw new CacheException(
                        "The auto-clean threashold must be a percentage (e.g. 50%) or a number of element. "
                                + res + " is not a correct value");
            return (this.cacheMaxSize * res) / 100;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.CacheManager#bind(java.lang.Object,
     * java.lang.Object)
     */
    public CacheEntry bind(Object id, Object object) throws CacheException {
        if (id == null || object == null) {
            throw new IllegalArgumentException(
                    "Impossible to add a cache entry with a null identifier or object");
        }

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "bind an element in the cache, id:" + id);
        }

        synchronized (this.cache) {
            this.clean();

            // checks that the cache is not full
            int size = this.cache.size();
            if (this.cacheMaxSize != NO_LIMIT) {
                if (size >= this.cacheMaxSize) {
                    if (this.cache.size() >= this.cacheMaxSize) {
                        // if the cache is still full, calls gc and waits for
                        // some time
                        this.forceClean();
                        size = this.cache.size();
                        if (size >= this.cacheMaxSize) {
                            // if the cache is really still full, log a warning
                            this.logger.log(BasicLevel.WARN,
                                    "The cache is still full after cleaning: number of exceeded entries is "
                                            + (size - this.cacheMaxSize));
                            // throw new CacheFullException();
                        }
                    }
                } else if ((size + 1) >= this.cacheThresHold && this.getBgcleaner() != null) {
                    this.getBgcleaner().backgroungCleanup(this.autoCleanSize);
                }
            }

            // checks that' id' is not already bound
            final WeakCacheEntry ref = this.cache.get(id);
            if (ref != null) {
                final FixableCacheEntry entry = (FixableCacheEntry) ref.get();
                if (entry != null && ref.ce != null) {
                    throw new CacheException("Identifier is already bound with "
                            + (entry.getCeObject() == object ? "the same" : "another")
                            + " object: \nid=" + id);
                }
            }

            // creates a new entry in the cache
            final FixableCacheEntry ce = this.cef.create(id, object);
            this.cache.put(id, new WeakCacheEntry(ce, this.queue));
            this.replacementManager.addForReplacement(ce);

            this.logger.log(BasicLevel.DEBUG, "Notify the binding to cache event listeners");
            final CacheEvent ev = new Event(CacheEvent.NOTIFY_BIND, ce, this.cache.size());
            for (final CacheEventListener cacheEventListener : this.cels.values()) {
                cacheEventListener.entryBound(ev);
            }

            return ce;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.CacheManager#lookup(java.lang.Object)
     */
    public CacheEntry lookup(Object id) {
        CacheEntry ce = null;

        if (id == null)
            throw new IllegalArgumentException(
                    "Impossible to find cache entry with a null identifier");

        synchronized (this.cache) {
            try {
                ce = this.get(id);
                if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                    this.logger.log(BasicLevel.DEBUG, "lookup: id=" + id);
                }
            } catch (final InvalidCacheEntryException e) {
                // do nothing
            }
        }

        return ce;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.CacheManager#fix(org.objectweb.dream.cache
     * .api.FixableCacheEntry)
     */
    public void fix(FixableCacheEntry _ce) throws CacheException {
        if (_ce == null) {
            throw new IllegalArgumentException("Impossible to fix a null cache entry");
        }

        synchronized (_ce) {
            final FixableCacheEntry ce = (FixableCacheEntry) this.get(_ce.getCeIdentifier());
            ce.fixCe();
            if (this.logger.isLoggable(BasicLevel.DEBUG))
                this.logger.log(BasicLevel.DEBUG, "Fix an element in the cache, id:"
                        + ce.getCeIdentifier() + " / count:" + ce.getCeFixCount());
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.CacheManager#fix(org.objectweb.dream.cache
     * .api.CacheEntryFilter)
     */
    public void fix(CacheEntryFilter filter) throws CacheException {
        synchronized (this.cache) { // to prevent lookup, bind in concurrence
            for (final WeakCacheEntry weakCacheEntry : this.cache.values()) {
                final CacheEntry ce = weakCacheEntry.ce;
                if (filter.accept(ce)) {
                    this.fix((FixableCacheEntry) ce);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.CacheManager#unfix(org.objectweb.dream.
     * cache.api.FixableCacheEntry)
     */
    public void unfix(FixableCacheEntry ce) throws CacheException {
        if (ce == null) {
            throw new IllegalArgumentException("Impossible to unfix a null cache entry");
        }

        synchronized (ce) {
            (ce).unfixCe();
            if (this.logger.isLoggable(BasicLevel.DEBUG))
                this.logger.log(BasicLevel.DEBUG, "UnFix an element in the cache, id:"
                        + ce.getCeIdentifier() + " / count:" + (ce).getCeFixCount());
        }
    }

    public void unfix(CacheEntryFilter filter) throws CacheException {
        synchronized (this.cache) { // to prevent lookup, bind in concurrence
            for (final WeakCacheEntry weakCacheEntry : this.cache.values()) {
                final CacheEntry ce = weakCacheEntry.ce;
                if (filter.accept(ce)) {
                    this.unfix((FixableCacheEntry) ce);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.CacheManager#touch(org.objectweb.dream.
     * cache.api.CacheEntry)
     */
    public void touch(CacheEntry ce) throws CacheException {
        if (ce == null)
            throw new IllegalArgumentException("Impossible to touch a null cache entry");

        synchronized (ce) {
            // Check that the entry exists in the cache
            this.replacementManager.adjustForReplacement((FixableCacheEntry) ce);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.CacheManager#unbindAll()
     */
    public void unbindAll() throws CacheException {
        this.unbindAll(true);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbindAll(boolean)
     */
    public void unbindAll(boolean force) throws CacheException {
        LoggerUtil.start(this.logger);

        synchronized (this.cache) { // to prevent lookup, bind in concurrence
            // clear the cache
            for (Object oid : this.cache.keySet()) {
                this._unbind(oid, force);
            }

            if (!force) {
                // In case of force == false, we run the GC in order to really
                // evict the entries
                this.forceClean();
                this.clean();
            }
        }

        LoggerUtil.end(this.logger);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbind(java.lang.Object,
     * boolean)
     */
    public boolean unbind(Object oid, boolean force) throws CacheException {
        synchronized (this.cache) { // to prevent lookup, bind in concurrence
            return this._unbind(oid, force);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.UnbindManager#unbind(org.objectweb.dream
     * .cache.api.CacheEntryFilter, boolean)
     */
    public Collection<Object> unbind(CacheEntryFilter filter, boolean force) throws CacheException {
        final Collection<Object> res = new ArrayList<Object>();
        synchronized (this.cache) { // to prevent lookup, bind in concurrence
            for (final WeakCacheEntry weakCacheEntry : this.cache.values()) {
                final CacheEntry ce = weakCacheEntry.ce;
                if (filter.accept(ce)) {
                    final Object oid = ce.getCeIdentifier();
                    if (this._unbind(ce, force)) {
                        res.add(oid);
                    }
                }
            }
        }
        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.UnbindManager#unbindAll(java.util.Collection
     * , boolean)
     */
    public Collection<Object> unbindAll(Collection<Object> oids, boolean force)
            throws CacheException {
        LoggerUtil.start(this.logger);

        ArrayList<Object> unbound = new ArrayList<Object>(oids.size());

        synchronized (this.cache) { // to prevent lookup, bind in concurrence

            if (this.cache.size() == 0) {
                unbound.addAll(oids);
                return unbound;
            }

            for (Object oid : oids) {
                if (!unbound.contains(oid)
                        && (this.lookup(oid) == null || this._unbind(oid, force))) {
                    unbound.add(oid);
                }
            }

            if (!force) {
                // In case of force == false, we run the GC in order to really
                // evict the entries
                this.forceClean();
                this.clean();
                if (this.cache.size() == 0) {
                    unbound = new ArrayList<Object>(oids);
                } else {
                    // Check if some entries has been removed
                    for (Object oid : oids) {
                        final WeakCacheEntry ref = this.cache.get(oid);
                        if (ref == null) {
                            if (!unbound.contains(oid)) {
                                unbound.add(oid);
                            }
                        }
                    }
                }
            }
        }

        LoggerUtil.end(this.logger);

        return unbound;
    }

    /**
     * Unbind the given Object.
     * 
     * @param oid
     * @param force
     *            Force the remove of the cache element
     * @return {@code true} if the cache element has been effectively removed
     * @throws CacheException
     */
    private boolean _unbind(Object oid, boolean force) throws CacheException {
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Unbind an element from the cache, id:" + oid);
        }

        this.clean();

        final WeakCacheEntry ref = this.cache.get(oid);
        if (ref == null) {
            return true;
        }
        final FixableCacheEntry ce = (FixableCacheEntry) ref.get();
        if (ce == null) {
            throw new InvalidCacheEntryException();
        }

        synchronized (ce) {
            if (ce.getCeFixCount() > 0) {
                return false;
            }
            // removes the strong reference to the real entry,
            // but does not remove it from' cache'
            ref.ce = null;
            if (force) {
                this.cache.remove(oid);
                final CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, this.cache.size());
                for (final CacheEventListener cacheEventListener : this.cels.values()) {
                    cacheEventListener.entryUnbound(ev);
                }
                this.replacementManager.removeForReplacement(oid);
            }
            return force;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbindUnfixed(boolean)
     */
    public Collection<Object> unbindUnfixed(boolean force) throws CacheException {
        throw new NotImplementedException();
        // return Collections.EMPTY_SET;
    }

    // ----------------------------------------------//

    private CacheEntry get(Object id) throws InvalidCacheEntryException {
        CacheEntry entry;

        synchronized (this.cache) {
            final WeakCacheEntry ref = this.cache.get(id);
            if (ref == null) {
                throw new InvalidCacheEntryException(id);
            }

            entry = ref.get();
            if (entry == null) {
                throw new InvalidCacheEntryException();
            }

            // Re-born the entry ;-)
            ref.ce = entry;
        }

        return entry;
    }

    /**
     * Removes all entries that have been garbage collected. Does not block.
     */
    private static final int NB_CALL_BEFORE_ACTUAL_CLEAN = 2;

    private int nbClean = 0;

    private void clean() throws CacheException {
        if (this.cacheMaxSize == NO_LIMIT || this.cache.size() < this.cacheThresHold) {
            if (this.nbClean < NB_CALL_BEFORE_ACTUAL_CLEAN) {
                this.nbClean++;
                return;
            }
        }

        this.nbClean = 0;

        WeakCacheEntry ref = (WeakCacheEntry) (this.queue.poll());
        while (ref != null) {
            if (ref.ce == null) {
                if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                    this.logger.log(BasicLevel.DEBUG,
                            "Remove from the cache during a clean the object with id:"
                                    + ref.ce.getCeIdentifier());
                }
                final Object oid = ref.ce.getCeIdentifier();
                this.cache.remove(oid);

                final CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, this.cache.size());
                for (final CacheEventListener cacheEventListener : this.cels.values()) {
                    cacheEventListener.entryUnbound(ev);
                }

                this.replacementManager.removeForReplacement(oid);
            }
            ref = (WeakCacheEntry) (this.queue.poll());
        }
    }

    /**
     * Calls the garbage collector and waits until an entry can be removed, or a
     * timeout period (TIMEOUTINMILLISECONDS milli-seconds) has expired.
     */
    private static final int TIMEOUTINMILLISECONDS = 50;

    private void forceClean() {
        this.logger.log(BasicLevel.DEBUG, "forceClean");

        System.gc();
        WeakCacheEntry ref = null;
        try {
            ref = (WeakCacheEntry) this.queue.remove(TIMEOUTINMILLISECONDS);
        } catch (final InterruptedException e) {
            // do nothing
        }
        if (ref != null && ref.ce == null) {
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG,
                        "Remove from the cache during a force clean the object with id:"
                                + ref.ce.getCeIdentifier());
            }

            final Object oid = ref.getIdentifier();
            this.cache.remove(oid);

            final CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, this.cache.size());
            for (final CacheEventListener cacheEntryEvent : this.cels.values()) {
                cacheEntryEvent.entryUnbound(ev);
            }

            this.replacementManager.removeForReplacement(oid);
            return;
        }

    }

    /**
     * The entries stored in 'cache' are weak references to the real entries.
     * Each weak entry also stores the id of the entry (this id is used in clean
     * and forceClean to remove the entry from 'cache')
     */
    private static class WeakCacheEntry extends WeakReference<CacheEntry> {
        private final Object id;
        // strong reference to the real entry: prevents it from being garbage
        // collected until the unbind method is called.
        public CacheEntry ce;

        public WeakCacheEntry(CacheEntry ce, ReferenceQueue<CacheEntry> queue) {
            super(ce, queue);
            this.id = ce.getCeIdentifier();
            this.ce = ce;
        }

        public Object getIdentifier() {
            return this.id;
        }
    }

    /**
     * This class is a basic implementation of the CacheEvent interface.
     */
    protected class Event implements CacheEvent {
        public int eventId = 0;
        public CacheEntry ce = null;
        public Object oid = null;
        public int cacheSize = -2;

        public Event(int eventid, Object oid, int cachesize) {
            this.eventId = eventid;
            this.oid = oid;
            this.cacheSize = cachesize;
        }

        public Event(int eventid, CacheEntry _ce, int cachesize) {
            this.eventId = eventid;
            this.ce = _ce;
            this.cacheSize = cachesize;
        }

        public int getEventId() {
            return this.eventId;
        }

        public Object getCeIdentifier() {
            return this.oid;
        }

        public CacheEntry getEntry() {
            return this.ce;
        }
    }

    /**
     * This class is a basic implementation of the CacheCapacityEvent interface.
     */
    protected class CapacityEvent implements CacheCapacityEvent {
        private final int eventId;
        private final int oldCacheSize;
        private final int cacheSize;

        public CapacityEvent(int eventid, int oldCacheSize, int cacheSize) {
            this.eventId = eventid;
            this.cacheSize = cacheSize;
            this.oldCacheSize = oldCacheSize;
        }

        public int getEventId() {
            return this.eventId;
        }

        public int getOldSize() {
            return this.oldCacheSize;
        }

        public int getSize() {
            return this.cacheSize;
        }
    }

    private BackgroundCleaner getBgcleaner() {
        if (this.bgcleaner == null && this.cacheThresHold > 0) {
            this.bgcleaner = new BackgroundCleaner(this.replacementManager, this.logger);
        }

        return this.bgcleaner;
    }
}
