/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.dream.cache.replacement.lib;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;

import org.objectweb.dream.cache.api.CacheCapacityEvent;
import org.objectweb.dream.cache.api.CacheCapacityEventListener;
import org.objectweb.dream.cache.api.CacheEntryFilter;
import org.objectweb.dream.cache.api.CacheException;
import org.objectweb.dream.cache.api.FixableCacheEntry;
import org.objectweb.dream.cache.api.UnbindManager;
import org.objectweb.dream.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.dream.cache.replacement.api.ReplacementManager;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is a common implementation of a ReplacementManager.
 * 
 * @author E.Bruneton
 * @see FIFOReplacementManager, LRUReplacementManager, MRUReplacementManager
 */
@DreamComponent
@Provides(interfaces = {
        @Interface(name = "replacement-manager", signature = ReplacementManager.class),
        @Interface(name = "cache-capacity-listeners", signature = CacheCapacityEventListener.class) })
public abstract class AbstractReplacementManager implements ReplacementManager,
        CacheCapacityEventListener {

    public final static String UNBIND_MANAGER_BINDING = "unbind-manager";

    @Service
    protected Component ref;

    @DreamMonolog
    protected Logger logger = null;

    @Requires(name = UNBIND_MANAGER_BINDING)
    protected UnbindManager unbindManager;

    /**
     * Sorted list of WeakReplaceableCacheEntry entries. Each entry must have a
     * unique age, so that this sorted set behaves as a sorted list.
     */
    protected Set<WeakReplaceableCacheEntry> entries;

    protected Map<Object, WeakReplaceableCacheEntry> oid2wentry;

    /**
     * Queue used to register the weak references stored in the entries list.
     */
    protected ReferenceQueue<ReplaceableCacheEntry> queue;

    public AbstractReplacementManager() {
        this.entries = new TreeSet<WeakReplaceableCacheEntry>();
        this.oid2wentry = new ConcurrentHashMap<Object, WeakReplaceableCacheEntry>();
        this.queue = new ReferenceQueue<ReplaceableCacheEntry>();
    }

    // IMPLEMENTATION OF THE LifeCycleControler INTERFACE //
    // ------------------------------------------------------//

    @DreamLifeCycle(on = DreamLifeCycleType.PREPARE_TO_STOP)
    public void stopFcHandler() throws IllegalLifeCycleException {
        LoggerUtil.call(this.logger);
        
        try {
            this.unbindAll(true);
        } catch (final CacheException e) {
            this.logger.log(BasicLevel.WARN, "The Replacement manager didn't stop correctly : "
                    + e.getMessage());
        }
    }

    // IMPLEMENTATION OF THE CacheCapacityEventListener INTERFACE //
    // --------------------------------------------------- //

    public void cacheResized(CacheCapacityEvent event) {
        final int size = event.getSize();
        if (size > 0) {
            synchronized (this.entries) {
                if (this.oid2wentry.size() == 0) {
                    this.oid2wentry = new HashMap<Object, WeakReplaceableCacheEntry>(size);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.replacement.api.ReplacementManager#
     * addForReplacement(org.objectweb.dream.cache.api.FixableCacheEntry)
     */
    public void addForReplacement(FixableCacheEntry entry) throws CacheException {
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "add the entry " + entry.getCeIdentifier());
        }

        final ReplaceableCacheEntry dentry = this.check(entry);
        synchronized (this.entries) {
            this.clean();
            WeakReplaceableCacheEntry we = this.oid2wentry.get(entry.getCeIdentifier());
            this.add(dentry);
            if (we != null) {
                this.entries.remove(we);
                we.updateEntry((ReplaceableCacheEntry) entry);
            } else {
                we = new WeakReplaceableCacheEntry(dentry, this.queue);
                this.oid2wentry.put(entry.getCeIdentifier(), we);
            }
            this.entries.add(we);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.replacement.api.ReplacementManager#
     * adjustForReplacement(org.objectweb.dream.cache.api.FixableCacheEntry)
     */
    public void adjustForReplacement(FixableCacheEntry entry) throws CacheException {
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "touch the entry " + entry.getCeIdentifier());
        }

        final ReplaceableCacheEntry dentry = this.check(entry);
        synchronized (this.entries) {
            this.clean();
            WeakReplaceableCacheEntry we = this.oid2wentry.get(entry.getCeIdentifier());
            this.touch(dentry);
            if (we != null) {
                this.entries.remove(we);
                we.updateEntry((ReplaceableCacheEntry) entry);
            } else {
                we = new WeakReplaceableCacheEntry(dentry, this.queue);
                this.oid2wentry.put(entry.getCeIdentifier(), we);
            }
            this.entries.add(we);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.replacement.api.ReplacementManager#forceFree
     * (int)
     */
    public int forceFree(int capacity) throws CacheException {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "forceFree: requested: " + capacity);
        }

        // elects entries which could be evicted
        final ArrayList<Object> oids = new ArrayList<Object>(capacity);
        int count = 0;
        synchronized (this.entries) {
            final Iterator<WeakReplaceableCacheEntry> i = this.entries.iterator();
            while (count < capacity && i.hasNext()) {
                final WeakReplaceableCacheEntry wentry = i.next();
                final ReplaceableCacheEntry entry = wentry.get();
                if (entry != null && entry.getCeFixCount() == 0) {
                    oids.add(entry.getCeIdentifier());
                    count++;
                }
            }
        }

        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "forceFree: try to unbound, oids=" + oids);
        }

        // Ask the eviction to the UnbindManager
        final Collection<Object> _oids = this.unbindManager.unbindAll(oids, false);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "forceFree: entries unbound, oid=" + _oids);
        }
        return _oids.size();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.objectweb.dream.cache.replacement.api.ReplacementManager#
     * removeForReplacement(java.lang.Object)
     */
    public void removeForReplacement(Object oid) {
        synchronized (this.entries) {
            final Object we = this.oid2wentry.remove(oid);
            if (we != null) {
                this.entries.remove(we);
            }
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.UnbindManager#unbind(org.objectweb.dream
     * .cache.api.CacheEntryFilter, boolean)
     */
    public Collection<Object> unbind(CacheEntryFilter filter, boolean force) throws CacheException {
        ArrayList<Object> oids;
        synchronized (this.entries) {
            if (this.entries.isEmpty()) {
                return Collections.emptySet();
            }
            oids = new ArrayList<Object>();
            for (final WeakReplaceableCacheEntry weakReplaceableCacheEntry : this.entries) {
                if (weakReplaceableCacheEntry != null) {
                    final ReplaceableCacheEntry ce = weakReplaceableCacheEntry.get();
                    if (ce != null && filter.accept(ce)) {
                        oids.add(weakReplaceableCacheEntry.getCeIdentifier());
                    }
                }
            }
        }
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "unbindAll(oids, " + force
                    + ") try to unbound, oids=" + oids);
        }
        // Ask the eviction to the UnbindManager
        final Collection<Object> res = this.unbindManager.unbindAll(oids, force);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + res);
        }
        return res;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.dream.cache.api.UnbindManager#unbindAll(java.util.Collection
     * , boolean)
     */
    public Collection<Object> unbindAll(Collection<Object> _oids, boolean force)
            throws CacheException {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        // Calculate the list of oid which match to unfixed entries
        ArrayList<Object> oids;
        synchronized (this.entries) {
            if (this.entries.isEmpty()) {
                return Collections.emptySet();
            }
            oids = new ArrayList<Object>();
            final Iterator<Object> it = _oids.iterator();
            while (it.hasNext()) {
                final Object oid = it.next();
                final WeakReplaceableCacheEntry wentry = this.oid2wentry.get(oid);
                if (wentry != null) {
                    oids.add(oid);
                }
            }
        }
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "unbindAll(oids, " + force
                    + ") try to unbound, oids=" + oids);
        }
        // Ask the eviction to the UnbindManager
        _oids = this.unbindManager.unbindAll(oids, force);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + _oids);
        }
        return _oids;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbindAll(boolean)
     */
    public void unbindAll(boolean force) throws CacheException {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        synchronized (this.entries) {
            if (this.entries.isEmpty()) {
                return;
            }
            this.entries.clear();
            this.oid2wentry.clear();
        }
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "unbindAll(" + force + ") try to unbound");
        }
        // Ask the eviction to the UnbindManager
        this.unbindManager.unbindAll(force);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "all entries unbound");
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbindUnfixed(boolean)
     */
    public Collection<Object> unbindUnfixed(boolean force) throws CacheException {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        // Calculate the list of oid matching to evictable entries
        ArrayList<Object> oids;
        synchronized (this.entries) {
            if (this.entries.isEmpty()) {
                return Collections.emptySet();
            }
            oids = new ArrayList<Object>();
            final Iterator<WeakReplaceableCacheEntry> it = this.entries.iterator();
            while (it.hasNext()) {
                final WeakReplaceableCacheEntry wentry = it.next();
                final ReplaceableCacheEntry ce = wentry.get();
                if (ce == null || ce.getCeFixCount() == 0) {
                    oids.add(wentry.getCeIdentifier());
                }
            }
        }
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "unbindAll(" + force + ") try to unbound, oids="
                    + oids);
        }
        // Ask the eviction to the UnbindManager
        final Collection<Object> _oids = this.unbindManager.unbindAll(oids, force);
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "entries unbound, oid=" + _oids);
        }
        return _oids;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.dream.cache.api.UnbindManager#unbind(java.lang.Object,
     * boolean)
     */
    public boolean unbind(Object oid, boolean force) throws CacheException {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        if (oid == null) {
            return false;
        }
        if (debug) {
            this.logger.log(BasicLevel.DEBUG, "unbind, oid=" + oid);
        }

        WeakReplaceableCacheEntry we = this.oid2wentry.get(oid);

        if (we != null) {
            return this.unbindManager.unbind(oid, force);
        } else {
            this.logger.log(BasicLevel.WARN, "The entry has not been found from the cache");
            return false;
        }
    }

    // ---------------------------------------------------//

    private ReplaceableCacheEntry check(FixableCacheEntry entry) {
        try {
            return (ReplaceableCacheEntry) entry;
        } catch (final ClassCastException e) {
            throw new IllegalArgumentException(
                    "This replacement manager works only with ReplaceableCacheEntry entries");
        }
    }

    protected abstract void add(ReplaceableCacheEntry entry);

    protected abstract void touch(ReplaceableCacheEntry entry);

    /**
     * Removes all entries that have been garbage collected. Does not block.
     */
    private void clean() {
        final boolean debug = this.logger.isLoggable(BasicLevel.DEBUG);
        while (true) {
            final WeakReplaceableCacheEntry entryRef = (WeakReplaceableCacheEntry) this.queue
                    .poll();
            if (entryRef != null) {
                if (debug) {
                    this.logger.log(BasicLevel.DEBUG, "auto cleanup, oid="
                            + entryRef.getCeIdentifier());
                }
                this.entries.remove(entryRef);
                this.oid2wentry.remove(entryRef.getCeIdentifier());
            } else {
                return;
            }
        }
    }

    /**
     * The entries stored in 'entries' are weak references to the real entries.
     * Each weak entry also stores the age of the real entry at the time it was
     * added in 'entries'.
     */
    private static class WeakReplaceableCacheEntry extends WeakReference<ReplaceableCacheEntry>
            implements Comparable<WeakReplaceableCacheEntry> {
        private long date;

        private Object oid;

        public WeakReplaceableCacheEntry(ReplaceableCacheEntry ce,
                ReferenceQueue<ReplaceableCacheEntry> queue) {
            super(ce, queue);
            this.oid = ce.getCeIdentifier();
            this.date = ce.getCeAge();
        }

        public Object getCeIdentifier() {
            return this.oid;
        }

        public int compareTo(WeakReplaceableCacheEntry wrce) {
            if (this.date == wrce.date) {
                return 0;
            } else {
                return this.date < wrce.date ? -1 : 1;
            }
        }

        @Override
        public boolean equals(Object o) {
            return this.date == ((WeakReplaceableCacheEntry) o).date;
        }

        public void updateEntry(ReplaceableCacheEntry ce) {
            this.oid = ce.getCeIdentifier();
            this.date = ce.getCeAge();
        }
    }
}
