/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.manager;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import org.objectweb.dream.control.activity.scheduler.InterruptTask;
import org.objectweb.dream.control.activity.scheduler.MultiThreadForwarderSchedulerImpl;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Scheduler manager that schedules each task with a
 * {@link MultiThreadForwarderSchedulerImpl} and initially no thread.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class MultiThreadPerTaskSchedulerManager
    extends
      AbstractSchedulerManager implements ThreadPoolManager, EndOfThread
{
  /**
   * Component reference
   */
  @Service
  Component                      weaveableC;

  /**
   * Logger used in concret implementation
   */
  @SuppressWarnings("unused")
  @DreamMonolog()
  protected Logger               logger;

  static final String            THREAD_PER_TASK_END_OF_THREAD_ITF_NAME = "tppt-end-of-thread";

  /**
   * High Level Scheduled Tasks that can have several thread executing the task.
   * The associated value contains the state of the task.
   */
  ConcurrentMap<Task, TaskState>           scheduledTask                          = new ConcurrentHashMap<Task, TaskState>();

  /**
   * Low Level tasks. The associated value contains the state of the
   * <b>high-level</b> task.
   */
  ConcurrentMap<Task, TaskState>           schedulerThread                        = new ConcurrentHashMap<Task, TaskState>();

  // ---------------------------------------------------------------------------
  // Attributes fields
  // ---------------------------------------------------------------------------

  long                           reviveTimeout                          = 2000;

  // ---------------------------------------------------------------------------
  // Implementation of the SchedulerManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see SchedulerManager#scheduleTask(Task, Map)
   * @return interface {@link ThreadPoolManager} to add thread to the task
   */
  public synchronized Object scheduleTask(Task task, Map hints)
      throws IllegalTaskException
  {
    LoggerUtil.start(logger);
    TaskState alreadyScheduledTaskState = scheduledTask.get(task);
    if (alreadyScheduledTaskState != null)
    {
      synchronized (alreadyScheduledTaskState)
      {
        // test if the task is going to stop
        if (alreadyScheduledTaskState.running == false){
          // wait for the end of all threads
          if (logger.isLoggable(BasicLevel.DEBUG)){
            logger.log(BasicLevel.DEBUG, "Waiting for the end of all threads for the task " + alreadyScheduledTaskState);
          }
          try
          {
            alreadyScheduledTaskState.wait();
          }
          catch (InterruptedException e)
          {
            Error.bug(logger);
          }
        } else {
          throw new IllegalTaskException(task, "Task already scheduled");
        }
      }
    }

    Scheduler scheduler = new MultiThreadForwarderSchedulerImpl();
    ThreadPoolManager ctrItf = addAndBindScheduler(task, scheduler);
    TaskState taskState = new TaskState(task, scheduler);
    scheduledTask.put(task, taskState);
    LoggerUtil.returnAfter(logger);
    return ctrItf;
  }

  /**
   * Add scheduler in the <code>activities</code> composite component and bind
   * it to the specified task.
   * 
   * @param task the task which will be linked to the </code>scheduler</code>
   * @param scheduler the scheduler to add and bind to the <code>task</code>
   * @return
   */
  private ThreadPoolManager addAndBindScheduler(Task task, Scheduler scheduler)
  {
    ThreadPoolManager ctrItf = null;
    try
    {
      contentControllerItf.addFcSubComponent(scheduler);
      ((BindingController) scheduler).bindFc(Task.ITF_NAME, task);
      ((BindingController) scheduler).bindFc(EndOfTask.ITF_NAME,
          contentControllerItf.getFcInternalInterface(EndOfTask.ITF_NAME));
      ((BindingController) scheduler).bindFc("logger", logger);
      ctrItf = (ThreadPoolManager) contentControllerItf
          .getFcInternalInterface(ThreadPoolManager.ITF_NAME);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
    return ctrItf;
  }

  /**
   * @see SchedulerManager#unscheduleTask(Task)
   */
  public synchronized void unscheduleTask(Task task)
      throws IllegalTaskException
  {
    LoggerUtil.start(logger);
    TaskState state = getTask(task);
    synchronized (state)
    {
      state.running = false;
      try
      {
        // EndofTask notified to task manager, just unbind scheduler
        ((BindingController) state.scheduler).unbindFc(Task.ITF_NAME);
        // wait for the EndOfThread to remove components.
      }
      catch (Exception e)
      {
        Error.bug(logger, e);
      }
      if (state.running == false && state.threads.isEmpty())
      {
        logger
            .log(BasicLevel.DEBUG, "Remove the task " + state + " from the scheduled tasks");
        scheduledTask.remove(state.task);
        state.notify();
      }
    }
    LoggerUtil.end(logger);
  }

  /**
   * @see InterruptTask#interruptTask(Task)
   */
  public synchronized void interruptTask(Task task) throws IllegalTaskException
  {
    LoggerUtil.start(logger);
    TaskState state = getTask(task);
    ((InterruptTask) state.scheduler).interruptTask(task);
    LoggerUtil.end(logger);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the EndOfThread interface
  // ---------------------------------------------------------------------------

  /**
   * @see EndOfThread#threadEnded(Task)
   */
  public boolean threadEnded(Task thread)
  {
    LoggerUtil.start(logger);
    TaskState state;
    state = schedulerThread.get(thread);
    unbindThread(thread, state);
    try
    {
      threadManagerItf.removeThread(thread);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
    synchronized (state)
    {
      if (state.running == false && state.threads.isEmpty())
      {
        logger
            .log(BasicLevel.DEBUG, "Remove the task from the scheduled tasks");
        scheduledTask.remove(state.task);
        state.notify();
      }
    }
    LoggerUtil.returnAfter(logger);
    return false;
  }

  /**
   * Unbind the thread from the related task. If the thread is not running and
   * there are no more thread executing the high-level task, the related
   * scheduler is removed.
   * 
   * @param thread the thread to unbind
   * @param state state of the high-level task
   */
  private void unbindThread(Task thread, TaskState state)
  {
    LoggerUtil.start(logger);
    synchronized (state)
    {
      if (state == null)
      {
        Error.bug(logger);
      }
      state.threads.remove(thread);
      schedulerThread.remove(thread);
      if (state.running == false && state.threads.isEmpty())
      {
        try
        {
          ((BindingController) thread).unbindFc(Scheduler.ITF_NAME);
          contentControllerItf.removeFcSubComponent(state.scheduler);
        }
        catch (Exception e)
        {
          Error.bug(logger, e);
        }
      }
    }
    LoggerUtil.end(logger);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ThreadPoolManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see ThreadPoolManager#getNbActiveThreads(Task)
   */
  public int getNbActiveThreads(Task task) throws IllegalTaskException
  {
    TaskState state = getTask(task);
    return state.threads.size();
  }

  /**
   * @see ThreadPoolManager#addThread(Task)
   */
  public synchronized void addThread(Task task)
      throws ThreadPoolOverflowException, IllegalTaskException
  {
    TaskState state = getTask(task);
    Task thread = null;
    thread = threadManagerItf.addThread();
    try
    {
      ((BindingController) thread).bindFc(Scheduler.ITF_NAME, state.scheduler);
      ((BindingController) thread).bindFc(EndOfThread.ITF_NAME,
          contentControllerItf
              .getFcInternalInterface(THREAD_PER_TASK_END_OF_THREAD_ITF_NAME));
      state.threads.add(thread);
      schedulerThread.put(thread, state);
      ((LifeCycleController) thread).startFc();
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
  }
  
  /**
   * Utilty method to get the state of a task
   * 
   * @param task the target task
   * @return the {@link TaskState} state of the task
   * @throws IllegalTaskException
   */
  private TaskState getTask(Task task) throws IllegalTaskException
  {
    TaskState state = scheduledTask.get(task);
    if (state == null)
    {
      throw new IllegalTaskException(task, "Unknown task");
    }
    return state;
  }

  // ---------------------------------------------------------------------------
  // inner classes
  // ---------------------------------------------------------------------------

  /**
   * State of the task containing following fields :
   * <ul>
   * <li>running : <code>true</code> if the task is currently running</li>
   * <li>task : the related task</li>
   * <li>scheduler : scheduler linked to the task</li>
   * <li>threads : pool of threads linked to the task</li>
   * </ul>
   */
  class TaskState
  {
    boolean   running = true;
    Task      task;
    Scheduler scheduler;
    Set<Task> threads = new HashSet<Task>();

    /**
     * Constructor
     * 
     * @param task related task
     * @param scheduler sheduler linked to the task
     */
    TaskState(Task task, Scheduler scheduler)
    {
      this.task = task;
      this.scheduler = scheduler;
    }
    
    @Override
    public String toString()
    {
      String res = ((NameController) task).getFcName() + "(state : " + running + "; threads count : " + threads.size();
      if (threads.size() > 0){
        res += "; threads : ";
        Iterator<Task> i = threads.iterator();
        while (i.hasNext()){
          res += i.next() + " ";
        }
      }
      res += ")";
      return res;
    }
  }
}
