/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.manager;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.scheduler.PeriodicSchedulerImpl;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * Scheduler manager for {@link PeriodicSchedulerImpl}.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class PeriodicSchedulerManager extends AbstractSchedulerManager
    implements
      EndOfThread
{
  /**
   * Component reference
   */
  @Service
  Component             weaveableC;

  /**
   * Logger used in concret implementation
   */
  @DreamMonolog()
  protected Logger      logger;

  /**
   * The hint to specify that a task is periodic
   */
  static final String   PERIOD_HINTS_KEY  = "period";

  /**
   * The scheduler for this kind of task. There is one scheduler (i.e. one
   * thread) for all periodic tasks. It is created when the first periodic task
   * is scheduled and cleared when the last task is unscheduled via the
   * {@link EndOfThread} interface
   */
  PeriodicSchedulerImpl periodicScheduler = null;

  /**
   * Task to schedule and their related period
   */
  Map<Task, String>     tasks             = new HashMap<Task, String>();
  
  /**
   * Id for each task
   */
  long                  nextTaskId        = 0;

  // ---------------------------------------------------------------------------
  // Implementation of the SchedulerManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see SchedulerManager#scheduleTask(Task, Map)
   */
  public Object scheduleTask(Task task, Map hints) throws IllegalTaskException
  {
    synchronized (tasks)
    {
      if (periodicScheduler == null)
      {
        try
        {
          periodicScheduler = new PeriodicSchedulerImpl();
          contentControllerItf.addFcSubComponent(periodicScheduler);
          Task thread = threadManagerItf.addThread();
          ((BindingController) periodicScheduler).bindFc(EndOfTask.ITF_NAME,
              contentControllerItf.getFcInternalInterface(EndOfTask.ITF_NAME));
          ((BindingController) thread).bindFc(Scheduler.ITF_NAME,
              periodicScheduler);
          ((BindingController) thread).bindFc(EndOfThread.ITF_NAME,
              this);
          ((BindingController) periodicScheduler).bindFc("logger", logger);
          ((LifeCycleController) thread).startFc();
        }
        catch (Exception e)
        {
          Error.bug(logger, e);
        }
      }
      Object o = hints.get(PERIOD_HINTS_KEY);
      if (o == null || !(o instanceof Number))
      {
        throw new IllegalTaskException(task, "Missing or invalid period");
      }
      String taskName = Task.ITF_NAME + (nextTaskId++);
      try
      {
        periodicScheduler.bindFc(taskName, task, o);
      }
      catch (Exception e)
      {
        Error.bug(logger, e);
      }
      tasks.put(task, taskName);
    }
    // no control on the task
    return null;
  }

  /**
   * @see SchedulerManager#unscheduleTask(Task)
   */
  public void unscheduleTask(Task task) throws IllegalTaskException
  {
    String taskName;
    synchronized (tasks)
    {
      taskName = tasks.remove(task);
    }
    if (taskName == null)
    {
      throw new IllegalTaskException(task, "unknown task");
    }
    try
    {
      periodicScheduler.unbindFc(taskName);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
  }

  /**
   * @see org.objectweb.dream.control.activity.scheduler.InterruptTask#interruptTask(Task)
   */
  public void interruptTask(Task task) throws IllegalTaskException
  {
    if (periodicScheduler == null)
    {
      throw new IllegalTaskException(task, "unknown task");
    }
    periodicScheduler.interruptTask(task);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the EndOfThread interface
  // ---------------------------------------------------------------------------

  /**
   * @see EndOfThread#threadEnded(Task)
   * @return <code>false</code> : the thread <b>must</b> never be restarted
   *         for this kind of scheduler.
   */
  public synchronized boolean threadEnded(Task thread)
  {
    if (periodicScheduler == null)
      Error.bug(logger, new RuntimeException("The periodic scheduler is null"));
    if (!tasks.isEmpty())
      Error
          .bug(
              logger,
              new RuntimeException(
                  "The periodic scheduler must not contains remaining task to schedule"));
    try
    {
      // don't unbind components, just remove them.
      contentControllerItf.removeFcSubComponent(this.periodicScheduler);
      threadManagerItf.removeThread(thread);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
    this.periodicScheduler = null;
    return false;
  }
}
