/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.manager;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.TaskStoppedListener;
import org.objectweb.dream.control.content.BasicContentController;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Basic task manager implementation
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class TaskManagerImpl implements TaskManager, EndOfTask
{

  /**
   * Component reference
   */
  @Service
  Component              weaveableC;

  /**
   * Logger used in concret implementation
   */
  @DreamMonolog()
  protected Logger       logger;

  /**
   * Map containing registered tasks
   */
  Map<Task, TaskState>   registeredTask                       = new HashMap<Task, TaskState>();

  // ---------------------------------------------------------------------------
  // client interfaces
  // ---------------------------------------------------------------------------

  /**
   * Interface used to schedule task : server interface of the meta-sheduler
   */
  @Requires(name = SchedulerManager.ITF_NAME)
  SchedulerManager       schedulerManagerItf;

  /** Name of the content controller client interface of this component. */
  static final String    ACTIVITY_CONTENT_CONTROLLER_ITF_NAME = "activities-content";
  @Requires(name = ACTIVITY_CONTENT_CONTROLLER_ITF_NAME)
  BasicContentController contentControllerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the TaskManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see TaskManager#registerTask(Task, Map)
   */
  public synchronized void registerTask(Task task, Map hints)
      throws IllegalTaskException
  {
    if (registeredTask.containsKey(task))
    {
      throw new IllegalTaskException(task, "Task already registered");
    }
    // add task as sub component
    try
    {
      contentControllerItf.addFcSubComponent(task);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }

    TaskState state = new TaskState(task, hints);
    registeredTask.put(task, state);

    // schedule task.
    Object o = schedulerManagerItf.scheduleTask(task, hints);
    task.registered(o);
  }

  /**
   * @see TaskManager#unregisterTask(Task)
   */
  public synchronized void unregisterTask(Task task)
      throws IllegalTaskException, InterruptedException
  {
    TaskState state = registeredTask.get(task);
    if (state == null)
    {
      throw new IllegalTaskException(task, "Unknown Task");
    }
    if (state.state == TaskState.RUNNING)
    {
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "the task " + task.toString()
            + " is currently running, going to interrupt it");
      }
      synchronized (state)
      {
        interruptTask(task, state);
        while (state.state != TaskState.STOPPED)
        {
          state.wait();
        }
      }
    }
    else
    {
      doUnregister(task);
    }
  }

  /**
   * @see TaskManager#interruptTask(Task, TaskStoppedListener)
   */
  public synchronized void interruptTask(Task task, TaskStoppedListener listener)
      throws IllegalTaskException
  {
    LoggerUtil.start(logger);
    TaskState state = registeredTask.get(task);
    if (state == null)
    {
      throw new IllegalTaskException(task, "Unknown Task");
    }
    if (state.state == TaskState.INTERRUPTED)
    {
      throw new IllegalTaskException(task, "task already interrupted");
    }
    state.state = TaskState.INTERRUPTED;
    state.listener = listener;
    schedulerManagerItf.interruptTask(task);
    LoggerUtil.end(logger);
  }

  /**
   * @see TaskManager#getTasks()
   */
  public Task[] getTasks()
  {
    Task[] tasks = new Task[registeredTask.size()];
    registeredTask.keySet().toArray(tasks);
    return tasks;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the EndOfTask interface
  // ---------------------------------------------------------------------------

  /**
   * @see EndOfTask#taskEnded(Task)
   */
  public void taskEnded(Task taskItf)
  {
    TaskState state = registeredTask.get(taskItf);
    if (state == null)
    {
      Error.bug(logger);
    }
    doUnregister(taskItf);
    if (state.listener != null)
    {
      state.listener.taskStopped(taskItf);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  /**
   * Utility Method to unregister a task
   * 
   * @param task task to unregister
   */
  private void doUnregister(Task task)
  {
    try
    {
      schedulerManagerItf.unscheduleTask(task);
      contentControllerItf.removeFcSubComponent(task);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
    registeredTask.remove(task);
    task.unregistered();
  }

  // ---------------------------------------------------------------------------
  // Inner class
  // ---------------------------------------------------------------------------

  class TaskState implements TaskStoppedListener
  {
    static final int    RUNNING     = 0;
    static final int    INTERRUPTED = 1;
    static final int    STOPPED     = 2;

    Task                task;
    Map                 hints;
    int                 state;
    TaskStoppedListener listener;

    /**
     * Constructor
     * 
     * @param task task related
     * @param hints hints linked to the task (example : period for a periodic
     *            task)
     */
    TaskState(Task task, Map hints)
    {
      this.task = task;
      this.hints = hints;
      state = RUNNING;
    }

    /**
     * @see TaskStoppedListener#taskStopped(Task)
     */
    public synchronized void taskStopped(Task task)
    {
      state = STOPPED;
      notifyAll();
    }
  }
}
