/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.control.activity.scheduler;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.objectweb.dream.control.activity.manager.EndOfTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This scheduler is only bound to one upper-level that it executes. The
 * <code>schedule</code> method only executes the upper-level task to which
 * this scheduler is bound. This scheduler can be multi threaded.
 */
public class MultiThreadForwarderSchedulerImpl
    implements
      Scheduler,
      InterruptTask,
      BindingController

{

  protected Set<Thread> executingThreads = new HashSet<Thread>();
  protected boolean     interrupted;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected Task        taskItf;
  protected EndOfTask   endOfTaskItf;

  // ---------------------------------------------------------------------------
  // Fields
  // ---------------------------------------------------------------------------

  protected Logger      logger;

  // ---------------------------------------------------------------------------
  // Implementation of the Scheduler interface
  // ---------------------------------------------------------------------------

  /**
   * @see Scheduler#schedule(int, Object)
   */
  public Object schedule(final int executionQuanta, final Object hints)
      throws InterruptedException
  {
    synchronized (this)
    {
      if (interrupted)
      {
        // task interrupted and executing thread not yet set,
        // activity manager has already been notified.
        return Task.STOP_EXECUTING;
      }
      executingThreads.add(Thread.currentThread());
    }
    Object o;
    do
    {
      try
      {
        logger.log(BasicLevel.DEBUG, "Execute Task.");
        o = taskItf.execute(hints);
        logger.log(BasicLevel.DEBUG, "End of task execution.");
      }
      catch (InterruptedException e)
      {
        if (!interrupted)
        {
          // not my interruption !
          throw e;
        }
        logger.log(BasicLevel.DEBUG, "InterruptedException caught.");
        o = Task.STOP_EXECUTING;
      }
      catch (Throwable e)
      {
        logger.log(BasicLevel.WARN,
            "Exception caught during execution of task: ", e);
        o = Task.STOP_EXECUTING;
      }
      if (interrupted)
      {
        o = Task.STOP_EXECUTING;
      }
    }
    while ((executionQuanta < 0) && (!Task.STOP_EXECUTING.equals(o)));
    synchronized (this)
    {
      executingThreads.remove(Thread.currentThread());
      if (interrupted && executingThreads.isEmpty())
      {
        endOfTaskItf.taskEnded(taskItf);
      }
    }
    return o;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the InterruptTask interface
  // ---------------------------------------------------------------------------

  /**
   * @see InterruptTask#interruptTask(Task)
   */
  public void interruptTask(Task task) throws IllegalTaskException
  {
    if (task != taskItf)
    {
      throw new IllegalTaskException(task);
    }
    synchronized (this)
    {
      interrupted = true;
      if (executingThreads.isEmpty())
      {
        endOfTaskItf.taskEnded(task);
      }
      else
      {
        Iterator<Thread> iter = executingThreads.iterator();
        while (iter.hasNext())
        {
          iter.next().interrupt();
        }
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see BindingController#bindFc(String, Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(Task.ITF_NAME))
    {
      taskItf = (Task) serverItf;
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      endOfTaskItf = (EndOfTask) serverItf;
    }
    else if (clientItfName.equals("logger"))
    {
      logger = (Logger) serverItf;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see BindingController#lookupFc(String)
   */
  public Object lookupFc(String clientItfName) throws NoSuchInterfaceException
  {
    if (clientItfName.equals(Task.ITF_NAME))
    {
      return taskItf;
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      return endOfTaskItf;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see BindingController#unbindFc(String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    if (clientItfName.equals(Task.ITF_NAME))
    {
      taskItf = null;
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      endOfTaskItf = null;
    }
    else if (clientItfName.equals("logger"))
    {
      logger = null;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Task.ITF_NAME, EndOfTask.ITF_NAME};
  }

}
