/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.task;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Provides a basic implementation of the
 * {@link org.objectweb.dream.control.activity.manager.TaskManager} interface.
 * <br>
 * <br>
 * <b>Requirements </b>
 * <ul>
 * <li>this mixin requires the {@link LoggableTaskMixin }mixin.</li>
 * <li>this mixin requires the
 * {@link org.objectweb.fractal.julia.control.binding.UseBindingControllerMixin}
 * mixin.</li>
 * </ul>
 */
public abstract class BasicTaskControllerMixin implements TaskController
{

  /** A map containing registration hints for each task. */
  protected Map<Task, Map<String, ? >> taskRegistrationHints;

  // ---------------------------------------------------------------------------
  // private constructor
  // ---------------------------------------------------------------------------

  /**
   * Private empty constructor
   */
  private BasicTaskControllerMixin()
  {
  }

  /**
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public void initFcController(final InitializationContext ic)
      throws InstantiationException
  {
    _super_initFcController(ic);

    taskRegistrationHints = Collections
        .synchronizedMap(new HashMap<Task, Map<String, ? >>());
  }

  // ---------------------------------------------------------------------------
  // Implementation of the TaskManagerController interface
  // ---------------------------------------------------------------------------

  /**
   * @see TaskController#getTasks()
   */
  public Task[] getTasks()
  {
    Task[] tasks = new Task[taskRegistrationHints.keySet().size()];
    return taskRegistrationHints.keySet().toArray(tasks);
  }

  /**
   * @see TaskController#addTask(Task, Map)
   */
  public void addTask(Task task, Map<String, Object > hints)
      throws IllegalTaskException
  {
    if (taskRegistrationHints.containsKey(task))
    {
      throw new IllegalTaskException(task,
          "The given task has already been added in the task controller", null);
    }
    taskRegistrationHints.put(task, hints);
  }

  /**
   * @see TaskController#removeTask(Task)
   */
  public void removeTask(Task task) throws NoSuchTaskException,
      IllegalTaskException
  {
    if (!taskRegistrationHints.containsKey(task))
    {
      throw new NoSuchTaskException(task);
    }
    _this_weaveableTCLogger.log(BasicLevel.DEBUG, "remove a task");
    try
    {
      taskRegistrationHints.remove(task);
    }
    catch (Exception e)
    {
      throw new IllegalTaskException(task, "Unable to unregister task", e);
    }
    _this_weaveableTCLogger.log(BasicLevel.DEBUG, "task removed");
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The <tt>weaveableTCLogger</tt> field required by this mixin. This field
   * is supposed to reference the logger of this controller.
   */
  public Logger            _this_weaveableTCLogger;

  /**
   * The <tt>weaveableBC</tt> field required by this mixin. This field is
   * supposed to reference the
   * {@link org.objectweb.fractal.api.control.BindingController}interface of
   * the component to which this controller object belongs.
   */
  public BindingController _this_weaveableBC;

  /**
   * The
   * {@link org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)}
   * method overridden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public abstract void _super_initFcController(InitializationContext ic)
      throws InstantiationException;
}
