/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.lifecycle;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.control.content.Util;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.LifeCycleCoordinator;

/**
 * Mixin that implements the coordination of asynchronous start.
 */
public abstract class AsyncStartLifeCycleCoordinatorMixin
    implements
      ASyncStarterCoordinator
{
  /**
   * Private constructor
   *
   */
  private AsyncStartLifeCycleCoordinatorMixin()
  {
  }

  // ---------------------------------------------------------------------------
  // Fields and methods added and overridden by the mixin class
  // ---------------------------------------------------------------------------

  protected Set<LifeCycleCoordinator> asyncStarting;
  protected IllegalLifeCycleException lifeCycleException;

  /**
   * The <tt>setFcState</tt> method overridden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleCoordinatorMixin#setFcState
   */
  public synchronized void setFcState(final boolean started)
      throws IllegalLifeCycleException
  {
    if (!started)
    {
      _super_setFcState(false);
    }
    else
    {
      asyncStarting = new HashSet<LifeCycleCoordinator>();
      lifeCycleException = null;
      // finds all the direct and indirect sub components of this component
      Component thisComponent;
      try
      {
        thisComponent = (Component) _this_weaveableC
            .getFcInterface("component");
      }
      catch (NoSuchInterfaceException e)
      {
        throw new ChainedIllegalLifeCycleException(e, _this_weaveableC,
            "Cannot set the life cycle state");
      }
      List allSubComponents = Util.getAllSubComponents(thisComponent);

      // sets the state of these components
      for (int i = 0; i < allSubComponents.size(); ++i)
      {
        Component c = (Component) allSubComponents.get(i);
        LifeCycleCoordinator lc;
        try
        {
          lc = (LifeCycleCoordinator) c.getFcInterface("lifecycle-controller");
        }
        catch (Exception e)
        {
          try
          {
            lc = (LifeCycleCoordinator) c
                .getFcInterface("/lifecycle-coordinator");
          }
          catch (NoSuchInterfaceException f)
          {
            continue;
          }
        }
        boolean asyncStart = false;
        try
        {
          NeedAsyncStartController asyncStartController = (NeedAsyncStartController) c
              .getFcInterface("need-async-start-controller");
          asyncStart = asyncStartController.getFcNeedAsyncStart();
        }
        catch (NoSuchInterfaceException e)
        {
          // ignore
        }
        if (asyncStart)
        {
          synchronized (asyncStarting)
          {
            asyncStarting.add(lc);
            new Thread(new AsyncStarter(lc, this)).start();
          }
        }
        else
        {
          lc.setFcStarted();
        }
      }
      synchronized (asyncStarting)
      {
        while (!asyncStarting.isEmpty() && lifeCycleException == null)
        {
          try
          {
            asyncStarting.wait();
          }
          catch (InterruptedException e)
          {
            throw new ChainedIllegalLifeCycleException(e, _this_weaveableC,
                "Unable to wait for the end of asynchronous starting");
          }
        }
      }
      if (lifeCycleException != null)
      {
        throw lifeCycleException;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ASyncStarterCoordinator interface
  // ---------------------------------------------------------------------------

  /**
   * @see ASyncStarterCoordinator#started(LifeCycleCoordinator,
   *      IllegalLifeCycleException)
   */
  public void started(LifeCycleCoordinator lc,
      IllegalLifeCycleException exception)
  {
    synchronized (asyncStarting)
    {
      if (asyncStarting.remove(lc)
          && (asyncStarting.isEmpty() || exception != null))
      {
        lifeCycleException = exception;
        asyncStarting.notify();
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The <tt>weaveableC</tt> field required by this mixin. This field is
   * supposed to reference the {@link Component} interface of the component to
   * which this controller object belongs.
   */

  public Component _this_weaveableC;

  /**
   * The <tt>setFcState</tt> method overridden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleCoordinatorMixin#setFcState
   */
  public abstract void _super_setFcState(boolean started)
      throws IllegalLifeCycleException;

}
