/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.lang.ref.Reference;
import java.util.Iterator;

import org.objectweb.dream.pool.Recyclable;

/**
 * Reference to a message created and held by a message {@link
 * MessageManagerType manager}.
 */
public abstract class Message implements Recyclable, ReferenceCounter {
    private final MessageManagerType owner;

    private int msgReferenceCounter = 0;

    private Reference<? extends Recyclable> recyclableReference;

    /**
     * Contructor
     * 
     * @param owner
     *      message {@link MessageManagerType manager} owner of the message
     */
    protected Message(final MessageManagerType owner) {
        this.owner = owner;
    }

    /**
     * Gets the owner of the message
     * 
     * @return message {@link MessageManagertype manager}
     */
    public final MessageManagerType getOwner() {
        return owner;
    }

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the Recyclable interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * Used only by message manager.
     * 
     * @see Recyclable#setReference(Reference)
     */
    public void setReference(final Reference<? extends Recyclable> reference) {
        this.recyclableReference = reference;
    }

    /**
     * Used only by message manager.
     * 
     * @see Recyclable#getReference()
     */
    public Reference<? extends Recyclable> getReference() {
        return recyclableReference;
    }

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the ReferenceCounter interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * Used only by message manager.
     * 
     * @see ReferenceCounter#decrementReferenceCounter()
     */
    public boolean decrementReferenceCounter() {
        return (--msgReferenceCounter) == 0;
    }

    /**
     * Used only by message manager.
     * 
     * @see ReferenceCounter#incrementReferenceCounter()
     */
    public void incrementReferenceCounter() {
        msgReferenceCounter++;
    }

    // ------------------------------------------------------------------------
    // ---
    // Abstract methods
    // ------------------------------------------------------------------------
    // ---

    /**
     * Tests whether the message message is empty or not
     */
    protected abstract boolean isEmpty();

    /**
     * Gets a chunk from its name
     * 
     * @param name
     *      name of the chunk
     * @return the chunk if it exist, <code>null</code> otherwise
     */
    protected abstract <T extends AbstractChunk<T>> T getChunk(String name);

    /**
     * Adds a chunk to a message
     * 
     * @param name
     *      name of the chunk
     * @param chunk
     *      chunk object
     */
    protected abstract <T extends AbstractChunk<T>> void addChunk(String name, T chunk);

    /**
     * Removes a chunk from a message
     * 
     * @param name
     *      name of the chunk to remove
     */
    protected abstract void removeChunk(String name);

    /**
     * Gets the names of chunk existing in the message
     * 
     * @return Iterator of names.
     */
    protected abstract Iterator<String> getNameIterator();

    /**
     * Adds a sub-message to this message
     * 
     * @param subMessage
     *      the message to add
     */
    protected abstract void addSubMessage(Message subMessage);

    /**
     * Tests whether a message is contained in this message
     * 
     * @param subMessage
     *      the potentiel sub-message
     * @return true if the message is a sub-message
     */
    protected abstract boolean isSubMessage(Message subMessage);

    /**
     * Removes a sub-message from this message
     * 
     * @param subMessage
     *      the message to remove
     * @return true if the sub-message was really remove
     */
    protected abstract boolean removeSubMessage(Message subMessage);

    /**
     * Gets the sub-messages
     * 
     * @return an {@link Iterator} of Message
     */
    protected abstract Iterator<Message> getSubMessageIterator();
}
