/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message.codec;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.ref.Reference;

import org.objectweb.dream.pool.Recyclable;

/**
 * An implementation of the {@link CodecInputOutput} interface that use byte
 * array.
 */
public class ByteArrayCodecInputOutput implements CodecInputOutput, Recyclable
{
  private static final byte[]              EMPTY_BYTE_ARRAY = new byte[0];
  protected Object                         input;
  protected Object                         output;
  private ByteArrayInputStream             byteArrayInputStream;
  private ByteArrayOutputStream            byteArrayOutputStream;

  private Reference< ? extends Recyclable> recyclableReference;

  /** Default constructor. */
  public ByteArrayCodecInputOutput()
  {
  }

  /**
   * @return the byte array contained in the output.
   */
  public byte[] getOutputByteArray()
  {
    if (byteArrayOutputStream == null)
    {
      return EMPTY_BYTE_ARRAY;
    }
    return byteArrayOutputStream.toByteArray();
  }

  /**
   * Initialize the input with the given byte array.
   * 
   * @param byteArray the input content.
   */
  public void setInputByteArray(byte[] byteArray)
  {
    byteArrayInputStream = new ByteArrayInputStream(byteArray);
    input = byteArrayInputStream;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the CodecInputOutput interface
  // ---------------------------------------------------------------------------

  /**
   * @see CodecInputOutput#getInput()
   */
  public Object getInput() throws IOException
  {
    return input;
  }

  /**
   * @see CodecInputOutput#setInput(Object)
   */
  public void setInput(Object input)
  {
    this.input = input;
  }

  /**
   * @see CodecInputOutput#getOutput()
   */
  public Object getOutput() throws IOException
  {
    if (output == null)
    {
      byteArrayOutputStream = new ByteArrayOutputStream();
      output = byteArrayOutputStream;
    }
    return output;
  }

  /**
   * @see CodecInputOutput#setOutput(Object)
   */
  public void setOutput(Object output)
  {
    this.output = output;
  }

  /**
   * @see CodecInputOutput#close()
   */
  public void close() throws IOException
  {
    if (byteArrayInputStream != null)
    {
      byteArrayInputStream.close();
    }
    if (byteArrayOutputStream != null)
    {
      byteArrayOutputStream.close();
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Recyclable interface
  // ---------------------------------------------------------------------------

  /**
   * @see Recyclable#recycle()
   */
  public void recycle()
  {
    input = null;
    output = null;
    byteArrayInputStream = null;
    byteArrayOutputStream = null;
  }

  /**
   * @see Recyclable#setReference(Reference)
   */
  public void setReference(Reference< ? extends Recyclable> reference)
  {
    this.recyclableReference = reference;
  }

  /**
   * @see Recyclable#getReference()
   */
  public Reference< ? extends Recyclable> getReference()
  {
    return recyclableReference;
  }

}
