/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message.codec;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Iterator;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * Codec helper implementation for composite messages. It scans the message
 * structure and recursively delegates to the client codec the encoding/decoding
 * of the message's chunks. This codec implementation requires a
 * {@link DataInput }as input and a {@link DataOutput }as output.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class MessageCodecHelperCompositeMessageDataStream extends AbstractDelegatingMessageCodec
        implements MessageCodec {

    @Requires(name = MessageManagerType.ITF_NAME)
    protected MessageManagerType messageManagerItf;

    /**
     * @see MessageCodec#encode(CodecInputOutput, Message)
     */
    public void encode(CodecInputOutput cio, Message message) throws IOException {
        DataOutput output = (DataOutput) cio.getOutput();
        doEncode(message, output, cio);
    }

    /**
     * @see MessageCodec#decode(CodecInputOutput)
     */
    public Message decode(CodecInputOutput cio) throws IOException {
        DataInput input = (DataInput) cio.getInput();
        return doDecode(input, cio);
    }

    // ---------------------------------------------------------------------------
    // Utility methods
    // ---------------------------------------------------------------------------

    /**
     * Decode recursively a message and its sub messages (if any). It delegates
     * to the client message codec to decode a single message.
     */
    protected Message doDecode(DataInput input, CodecInputOutput cio) throws IOException {
        Message message = delegateMessageCodecItf.decode(cio);
        boolean hasSubMessage = input.readBoolean();
        // read unnamed sub message.
        while (hasSubMessage) {
            Message subMessage = doDecode(input, cio);
            messageManagerItf.addSubMessage(message, subMessage);
            hasSubMessage = input.readBoolean();
        }
        return message;
    }

    /**
     * Encode recursively a message and its sub messages (if any). It delegates
     * to the client message codec to encode a single message.
     */
    protected void doEncode(Message message, DataOutput output, CodecInputOutput cio)
            throws IOException {
        delegateMessageCodecItf.encode(cio, message);
        Iterator<Message> iterator = messageManagerItf.getSubMessageIterator(message);
        while (iterator.hasNext()) {
            output.writeBoolean(true);
            Message subMessage = iterator.next();
            doEncode(subMessage, output, cio);
        }
        output.writeBoolean(false);
    }
}
