/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.message.codec.string;

import java.io.IOException;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.ref.Reference;

import org.objectweb.dream.message.codec.CodecInputOutput;
import org.objectweb.dream.pool.Recyclable;

/**
 * An implementation of the {@link CodecInputOutput} interface that use String
 * format
 */
public class StringCodecInputOutput implements CodecInputOutput, Recyclable {
    private static final String EMPTY_STRING = "";

    /**
     * Stream used to deserialze the stream of objects from the XML
     */
    protected Object input;

    /**
     * Stream used to serialize a stream of objects to XML
     */
    protected Object output;

    /**
     * Writer used by the xstreamOutputStream. The output XML string can be
     * retrieved with this writer.
     */
    protected StringWriter stringWriter;

    /**
     * Reader used by the xstreamInputStream. The input XML string can be set
     * with this reader.
     */
    protected StringReader stringReader;

    /**
     * Reference used by the pool of objets
     */
    protected Reference<? extends Recyclable> recyclableReference;

    // ---------------------------------------------------------------------------
    // XmlCodecInputOutput specific methods
    // ---------------------------------------------------------------------------

    /**
     * @return the result string contained in the output.
     */
    public String getOutputString() {
        if (this.stringWriter == null) {
            return EMPTY_STRING;
        }
        return this.stringWriter.toString();
    }

    /**
     * Initialize the input with the string.
     * 
     * @param byteArray
     *            the input content.
     */
    public void setInputString(String string) {
        this.stringReader = new StringReader(string);
        this.input = this.stringReader;
    }

    // ---------------------------------------------------------------------------
    // Implementation of the CodecInputOutput interface
    // ---------------------------------------------------------------------------

    /**
     * @see org.objectweb.dream.message.codec.CodecInputOutput#close()
     */
    public void close() throws IOException {
        if (this.stringReader != null) {
            this.stringReader.close();
        }
        if (this.stringWriter != null) {
            this.stringWriter.close();
        }
    }

    /**
     * @see org.objectweb.dream.message.codec.CodecInputOutput#getInput()
     */
    public Object getInput() throws IOException {
        return this.input;
    }

    /**
     * Set the XML {@link InputStream} of this codec. This method is useful to
     * have a direct access to the read string.
     * 
     * @param output
     *            a {@link StringWriter} used by the XMLOutputStream
     */
    public void setInput(Object input) {
        this.input = input;
    }

    /**
     * @see org.objectweb.dream.message.codec.CodecInputOutput#getOutput()
     */
    public Object getOutput() throws IOException {
        if (this.output == null) {
            stringWriter = new StringWriter();
            this.output = stringWriter;
        }
        return this.output;
    }

    /**
     * Set the XML {@link OutputStream} of this codec.
     */
    public void setOutput(Object output) {
        this.output = output;
    }

    // ---------------------------------------------------------------------------
    // Implementation of the Recyclable interface
    // ---------------------------------------------------------------------------

    public void recycle() {
        input = null;
        output = null;
        stringReader = null;
        stringWriter = null;
    }

    /**
     * @see Recyclable#setReference(Reference)
     */
    public void setReference(Reference<? extends Recyclable> reference) {
        this.recyclableReference = reference;
    }

    /**
     * @see Recyclable#getReference()
     */
    public Reference<? extends Recyclable> getReference() {
        return recyclableReference;
    }
}
