/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.pool;

import java.util.HashSet;
import java.util.Set;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * ObjectPoolItf that checks that objects are recycled only once. This component
 * can be used for debug purpose.
 * 
 * @see org.objectweb.dream.pool.ObjectPool
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class ObjectPoolDebug implements ObjectPoolItf
{
  /**
   * Component reference
   */
  @Service
  Component weaveableC;
  
  /**
   * Logger used in concret implementation
   */ 
  @DreamMonolog()
  protected Logger                   logger;

  private int                   newInstancesCount = 0;
  /** The pool of objects. */
  private Recyclable            pool[];

  /** The number of elements stored in the pool. */
  private int                   elementCount      = 0;

  /** The capacity of the pool. */
  @SuppressWarnings("unused")
  @Attribute(argument = "capacity")
  private int                   capacity;

  private final Object          lock              = new Object();

  Set<Recyclable>               recycledObjects   = new HashSet<Recyclable>();
  Set<Recyclable>               createdObjects    = new HashSet<Recyclable>();

  // ---------------------------------------------------------------------------
  // Client interface
  // ---------------------------------------------------------------------------

  private RecyclableFactoryType recyclableFactoryItf;

  // ---------------------------------------------------------------------------
  // Implementation of the ObjectPoolItf interface
  // ---------------------------------------------------------------------------

  /**
   * @see ObjectPoolItf#newInstance()
   */
  public Recyclable newInstance()
  {
    Recyclable obj = null;
    synchronized (lock)
    {
      if (elementCount == 0)
      {
        if (logger.isLoggable(BasicLevel.INFO))
        {
          newInstancesCount++;
          logger.log(BasicLevel.INFO,
              "Creates new instance. created instances = " + newInstancesCount);
        }
        obj = recyclableFactoryItf.newInstance();
      }
      else
      {
        elementCount -= 1;
        obj = pool[elementCount];
        pool[elementCount] = null;
        if (!recycledObjects.remove(obj))
        {
          Error.bug(logger);
        }
      }
      if (!createdObjects.add(obj))
      {
        Error.bug(logger);
      }
    }
    return obj;
  }

  /**
   * @see ObjectPoolItf#recycleInstance(Recyclable)
   */
  public void recycleInstance(Recyclable recyclable)
  {
    if (recyclable != null)
    {
      recyclable.recycle();
      synchronized (lock)
      {
        if (!createdObjects.contains(recyclable))
        {
          logger.log(BasicLevel.ERROR,
              "Recycling an object that has not been created by this pool",
              new Exception());
          Error.bug(logger);
        }

        if (!createdObjects.remove(recyclable))
        {
          Error.bug(logger);
        }

        if (recycledObjects.contains(recyclable))
        {
          logger.log(BasicLevel.ERROR, "Recycling an already recycled object",
              new Exception());
          Error.bug(logger);
        }

        if (!recycledObjects.add(recyclable))
        {
          Error.bug(logger);
        }

        if (elementCount == pool.length)
        {
          if (logger.isLoggable(BasicLevel.INFO))
          {
            logger.log(BasicLevel.INFO, "Pool Full created instances = "
                + newInstancesCount);
          }
          return;
        }
        pool[elementCount] = recyclable;
        elementCount += 1;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Override the generated implementation of ObjectPoolAttributeControllerOld
  // interface
  // ---------------------------------------------------------------------------

  /**
   * Set the capacity of the pool
   * 
   * @param capacity new capacity of the pool
   */
  public void setCapacity(int capacity)
  {
    // TODO try to re-use previous instance
    synchronized (lock)
    {
      this.capacity = capacity;
      pool = new Recyclable[capacity];
      elementCount = 0;
    }
  }
}
